import time
import sys
sys.path.append('C:\Program Files (x86)\Keysight\SD1\Libraries\Python')
import keysightSD1
import keysight_hvi as kthvi
import numpy
from threading import Thread
import matplotlib
import matplotlib.pyplot as plt

"""
Pathwave Test Sync Executive - Synchronous Mixed-Signal Measurements using M3xxxA PXI Instruments

Copyright Keysight Technologies 2020

In this application a digitizer performs sequenced acquisition of heterogeneous signals being generated by multiple AWGs.
More specifically AWG_1 generates a train of RF pulses, the other AWGs output a previously queued arbitrary waveform.

The showcased HVI API 2020 functionalities are:
* Scalability to an arbitrary number of M3xxxA PXI instruments
* Off-shelf inter-instrument synchronization capabilities
* HVI Synchronized Multi-Sequence Block to synchronously parallel-execute sequences in multiple instruments
* HVI Synchronized While Statement to run synchronized loops over multiple instruments
* HVI Local Flow Control: Wait Statement, IF Statement, WHILE Statement

The Python code that follows: 
1. Opens an arbitrary number of M3xxxA instruments (real hardware or simulation mode) 
2. Defines HVI resources (engines, chassis, chassis interconnections, trigger, action, events, HVI triggering resources, etc)
3. Programs HVI sequences by adding the necessary statements and instructions in each AWG's sequence
4. Compiles HVI, loads it to HW and executes it

For further info please check the programming example titled "Pathwave Test Sync Executive - Synchronous Mixed-Signal Measurements using M3xxxA PXI Instruments"
"""
######################################
# Application Configuration 
######################################

class ApplicationConfig:
    """
    Defines module descriptors, configuration options and names of HVI engines, actions, triggers
    """
    def __init__(self):
        """
        Define PXI instruments options
        """
        # Define options to open the instruments. 
        # Complete option list can be found in the SD1 3.x User Guide
        self.options = 'channelNumbering=keysight'
        # Simulated HW Option
        self.hardware_simulated = True
        """ 
        Define names of HVI engines, actions, triggers, registers
        """
        # HVI engines
        self.awg_engine = "AWG Engine"
        self.rf_gen_engine = "RF Generator Engine"
        self.dig_engine = "Digitizer Engine"
        # HVI actions
        self.awg_trigger = "AWG_Trigger"
        self.daq_trigger = "DAQ_Trigger"
        # HVI triggers
        self.fp_trigger = "FP Trigger"
        # HVI registers
        self.iteration_counter = "Iteration Counter"
        self.pulse_counter = "Pulse Counter"
        self.queue_reg = "Queue Reg"
        self.reg_wfm_A = "Wfm A"
        self.reg_wfm_B = "Wfm B"
        self.hvi_done = "HVI Done"
        """
        Define HW Platform
        """
        # Define module descriptors below with your instruments information1
        self.digitizer_descriptor = ModuleDescriptor('M3102A', 1, 9, self.options, self.dig_engine)
        self.rf_gen_descriptor = ModuleDescriptor('M3202A', 1, 8, self.options, self.rf_gen_engine) # instrument to be used as an RF Pulse Gen.
        self.awg_descriptors = [ModuleDescriptor('M3202A', 1, 7, self.options, "")]
        self.define_awg_engine_names() # assigns HVI engine names to awg_descriptors
        self.module_descriptors = [] # combines all previous descriptors
        self.define_module_descriptors()
        
        # Define list of chassis containing instruments to be used by HVI
        self.chassis_list = [1]
        
        # Multi-chassis setup
        # In case of multiple chassis, chassis PXI lines need to be shared using M9031 PXI modules. 
        # See programming example documentation for more details. For example in a three-chassis setup:
        # self.M9031_descriptors = [
        #     M9031Descriptor(1, 11, 2, 10),
        #     M9031Descriptor(2, 16, 3, 9)]
        # In case of single-chassis setup please leave the list empty: M9031_descriptors = [] 
        self.M9031_descriptors = []
        
        # Assign triggers to HVI object to be used for HVI-managed synchronization, data sharing, etc
        # NOTE: In a multi-chassis setup ALL the PXI lines listed below need to be shared among each M9031 board pair by means of SMB cable connections
        self.pxi_sync_trigger_resources = [
            kthvi.TriggerResourceId.PXI_TRIGGER5, 
            kthvi.TriggerResourceId.PXI_TRIGGER6, 
            kthvi.TriggerResourceId.PXI_TRIGGER7]
        """
        AWG Parameters
        """
        # AWG settings for all channels
        self.sync_mode = keysightSD1.SD_SyncModes.SYNC_NONE
        self.queue_mode = keysightSD1.SD_QueueMode.ONE_SHOT
        self.awg_mode = keysightSD1.SD_Waveshapes.AOU_AWG
        self.start_delay = 0 # x10 [ns]
        self.awg_prescaler = 0
        self.wfm_A_cycles = 3
        self.wfm_B_cycles = 2
        self.amplitude = 1
        self.wfm_A = 0
        self.wfm_B = 1
        # Trigger settings
        self.awg_trigger_mode = keysightSD1.SD_TriggerModes.SWHVITRIG
        # Latency values for M3202A AWGqueueWfm() [ns]
        # Latencies depend on AWG FPGA FW. Check M3xxxA User Guide for further info
        self.queue_wfm_latency = 100 # [ns] Minimum start delay necessary to execute an AWGqueueWfm() instruction
        self.awg_trigger_latency = 2000 # [ns] Minimum latency necessary between an AWGqueueWfm() instruction and an AWGtrigger action.
        self.wfm_length = 100 # [ns]
        """ 
        RF pulse generator parameters
        """
        self.all_ch_mask = 0xF # binary mask defining which channels to use
        self.offset = 0 # [V]
        self.frequency = 10e6 # [Hz]
        self.num_loops = 3 # sync while loops
        self.num_pulses = 5
        self.ON_value = 1.0 # [V]
        self.OFF_value = 0.0 # [V]
        self.n_AWG = 1 # channel number to be used as RF Gen
        self.pulse_ontime = 200 # [ns] 
        self.pulse_delay = 100 # [ns]
        """ 
        Digitizer parameters
        """
        self.dig_sampling_time = 2 # [ns] 1/sample_rate, sample_rate = 500 MSa/s for Digitizer M3102A 
        self.acquisition_points_per_cycle = 1500
        self.dig_prescaler = 0
        self.fullscale = 2 # [V] enter x Volts to set the full scale to [-x, x] Volts
        self.acquisition_time_per_cycle = self.acquisition_points_per_cycle*self.dig_sampling_time
        self.dig_num_cycles = self.num_loops #insert -1 for infinite cycles
        self.acquisition_points = int(self.acquisition_points_per_cycle*self.num_loops)
        self.acquisition_delay = 150 # x2 [ns]
        self.dig_trigger_mode = keysightSD1.SD_TriggerModes.SWHVITRIG
        self.mask = self.all_ch_mask
    
    def define_awg_engine_names(self):
        " Define names for all AWG HVI engines"

        # Assign AWG engine names to AWG1-AWGN in case more than AWG are used
        for index in range(len(self.awg_descriptors)):
            self.awg_descriptors[index].engine_name = self.awg_engine + str(index)

        pass
    
    def define_module_descriptors(self):
        """
        Combines all modules descriptor into a single module_descriptors array
        """
        # Compose array of module descriptors
        for awg in self.awg_descriptors:
            self.module_descriptors.append(awg)
        self.module_descriptors.append(self.rf_gen_descriptor)
        self.module_descriptors.append(self.digitizer_descriptor)

        pass

######################################
# Main Program
######################################

def main():
    ######################################################
    # Open PXI instruments, configure AWGs and digitizers
    ######################################################
    # Load configuration
    config = ApplicationConfig()

    # Checks flags defining if the code executes on HW or Simulation mode
    if config.hardware_simulated:
        print("Code running in Simulation Mode")

    # Open instruments
    module_dict = open_modules(config)
    
    # Configure Digitizer
    configure_digitizer(config, module_dict[config.dig_engine].instrument)
    # Configure RF Gen.
    configure_rfgen(config, module_dict[config.rf_gen_engine].instrument)
    # Configure AWGs 1:N
    for awg in config.awg_descriptors:
        awg_module = module_dict[awg.engine_name]
        configure_awg(config, awg_module.instrument, awg_module.num_channels)

    ######################################
    # System definition
    ######################################
    my_system = kthvi.SystemDefinition("MySystem") 
    # Define your system, HW platform and HVI resources
    define_hvi_resources(my_system, module_dict, config)
    
    ######################################
    # Program HVI sequences
    ######################################
    print("Programming the HVI sequences...")
    # Create sequencer object
    sequencer = kthvi.Sequencer("MySequencer", my_system)
    # Program HVI sequences using the sequencer object
    program_mixed_sig_meas_sequence(sequencer, module_dict, config)
    # Export the programmed sequence to text
    export_hvi_seqeunces(sequencer)
    
    ########################################
    # Compile, Load to HW, Run HVI Sequence
    ########################################
    # Compile HVI sequences
    hvi = sequencer.compile()
    print(hvi.compile_status.to_string())
    print("HVI Compiled")
    print("This HVI application needs to reserve {} PXI trigger resources to execute".format(len(hvi.compile_status.sync_resources)))
    
    # Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
    hvi.load_to_hw()    
    print("HVI Loaded to HW")

    # Execute HVI in non-blocking mode 
    # This mode allows SW execution to interact with HVI execution
    hvi.run(hvi.no_wait)
    print("HVI Running...")

    # Interact with HVI execution
    if not config.hardware_simulated:
        interact_with_hvi(hvi, module_dict, config)

    # Release HW resources once HVI execution is completed
    hvi.release_hw()
    print("Releasing HW...")

    # Close all modules at the end of the execution
    for module in module_dict.values():
        module.instrument.close()
    print("Modules closed\n")

########################################################
# 2nd Level: Functions to Define, Program, Execute HVI
#######################################################

def define_hvi_resources(sys_def, module_dict, config):
    """
    Configures all the necessary resources for the HVI application to execute: HW platform, engines, actions, triggers, etc. 
    """
    # Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
    define_hw_platform(sys_def, config)

    # Define all the HVI engines to be included in the HVI
    define_hvi_engines(sys_def, module_dict)
    
    # Define list of actions to be executed
    define_hvi_actions(sys_def, module_dict, config)

    # Defines the trigger resources
    define_hvi_triggers(sys_def, module_dict, config)


def define_hw_platform(sys_def, config):
    """
     Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks 
    """
    # Add chassis resources 
    # For multi-chassis setup details see programming example documentation
    for chassis_number in config.chassis_list:
        if config.hardware_simulated:
            sys_def.chassis.add_with_options(chassis_number, 'Simulate=True,DriverSetup=model=M9018B,NoDriver=True')
        else:
            sys_def.chassis.add(chassis_number)

    # Add M9031 modules for multi-chassis setups
    if config.M9031_descriptors:
        interconnects = sys_def.interconnects
        for descriptor in config.M9031_descriptors:
            interconnects.add_M9031_modules(descriptor.chassis_1, descriptor.slot_1, descriptor.chassis_2, descriptor.slot_2)

    # Assign the defined PXI trigger resources
    sys_def.sync_resources = config.pxi_sync_trigger_resources
    
    # Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
    # Use the code line below if you want the application to be in sync with the 10 MHz clock
    sys_def.non_hvi_core_clocks = [10e6]


def define_hvi_engines(sys_def, module_dict):
    """
    Define all the HVI engines to be included in the HVI
    """
    # For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
    for engine_name in module_dict.keys():
        sys_def.engines.add(module_dict[engine_name].instrument.hvi.engines.main_engine, engine_name)


def define_hvi_actions(sys_def, module_dict, config):
    """ 
    Defines AWG trigger actions for each module, to be executed by the "action execute" instruction in the HVI sequence
    Create a list of AWG trigger actions for each AWG module. The list depends on the number of channels 
    """

    # For each AWG, define the list of HVI Actions to be executed and add such list to its own HVI Action Collection
    for engine_name in module_dict.keys():
        for ch_index in range(1, module_dict[engine_name].num_channels + 1):

            # Actions need to be added to the engine's action list so that they can be executed
            if engine_name == config.dig_engine:
                action_name = config.daq_trigger + str(ch_index) # arbitrary user-defined name
                instrument_action = "daq{}_trigger".format(ch_index) # name decided by instrument API
            else:
                action_name = config.awg_trigger + str(ch_index) # arbitrary user-defined name
                instrument_action = "awg{}_trigger".format(ch_index) # name decided by instrument API
            
            action_id = getattr(module_dict[engine_name].instrument.hvi.actions, instrument_action)
            sys_def.engines[engine_name].actions.add(action_id, action_name)


def define_hvi_triggers(sys_def, module_dict, config):
    " Defines the FP trigger to be used as a wait condition by the digitizer "
    
    # Add to the HVI Trigger Collection of each HVI Engine the FP Trigger object of that same instrument
    fp_trigger_id = module_dict[config.dig_engine].instrument.hvi.triggers.front_panel_1
    fp_trigger = sys_def.engines[config.dig_engine].triggers.add(fp_trigger_id, config.fp_trigger)

    # Trigger configuration
    # NOTE: Trigger to be used as WaitEvent conditions must be configured as kthvi.Direction.INPUT
    # DriveMode (e.g. PushPull/OpenDrain) is defined by the instrument and cannot be changed by the user
    fp_trigger.config.direction = kthvi.Direction.INPUT
    fp_trigger.config.polarity = kthvi.Polarity.ACTIVE_HIGH
    fp_trigger.config.hw_routing_delay = 0
    fp_trigger.config.trigger_mode = kthvi.TriggerMode.LEVEL


def program_mixed_sig_meas_sequence(sequencer, module_dict, config):
    """
    This method programs the HVI sequence of this application.
    Different HVI statements are encapsulated as much as possible in separated SW methods to help users visualize
    the programmed HVI sequences.
    The programming example documentation on www.keysight.com contains an HVI diagram that graphically represents the programmed HVI sequence.
    """
    # Define registers within the scope of the outmost sync sequence
    define_registers(sequencer, module_dict, config)

    # Define sync while condition
    iteration_counter = sequencer.sync_sequence.scopes[config.dig_engine].registers[config.iteration_counter]
    sync_while_condition = kthvi.Condition.register_comparison(iteration_counter, kthvi.ComparisonOperator.LESS_THAN, config.num_loops)
    # Add Sync While Statement
    sync_while = sequencer.sync_sequence.add_sync_while("Sequenced Acquisition Loop", 90, sync_while_condition)
    # Program Sequenced Acquisition Loops
    program_sequenced_meas_loop(sync_while.sync_sequence, module_dict, config)
    
    # Add 3rd Sync Multi-Sequence Block
    sync_block = sequencer.sync_sequence.add_sync_multi_sequence_block("Execution Completed", 260)
    # Program execution completed
    program_execution_completed(sync_block, config)


def define_registers(sequencer, module_dict, config):
    " Defines all registers for each HVI engine in the scope af the global sync sequence "
    
    # Digitizer registers
    iteration_counter = sequencer.sync_sequence.scopes[config.dig_engine].registers.add(config.iteration_counter, kthvi.RegisterSize.SHORT)
    iteration_counter.initial_value = 0
    hvi_done = sequencer.sync_sequence.scopes[config.dig_engine].registers.add(config.hvi_done, kthvi.RegisterSize.SHORT)
    hvi_done.initial_value = 0
    
    # RF Gen registers
    pulse_counter = sequencer.sync_sequence.scopes[config.rf_gen_engine].registers.add(config.pulse_counter, kthvi.RegisterSize.SHORT)
    pulse_counter.initial_value = 0
    
    # AWG 1:N Registers
    for engine_name in module_dict.keys():
        if engine_name!=config.rf_gen_engine and engine_name!=config.dig_engine:
            queue_reg = sequencer.sync_sequence.scopes[engine_name].registers.add(config.queue_reg, kthvi.RegisterSize.SHORT)
            queue_reg.initial_value = 0
            reg_wfm_A = sequencer.sync_sequence.scopes[engine_name].registers.add(config.reg_wfm_A, kthvi.RegisterSize.SHORT) 
            reg_wfm_A.initial_value = config.wfm_A
            reg_wfm_B = sequencer.sync_sequence.scopes[engine_name].registers.add(config.reg_wfm_B, kthvi.RegisterSize.SHORT) 
            reg_wfm_B.initial_value = config.wfm_B


def program_sequenced_meas_loop(sync_sequence, module_dict, config):
    "Programs the outmost Sync While statement of the HVI Sync Sequence"

    # Add 1st Sync Multi-Sequence Block to the Sync While sequence
    sync_block_1 = sync_sequence.add_sync_multi_sequence_block("Loop Initialization", 270)
    # Program the "Wait for External Trigger" Sync Multi-Sequence Block
    program_loop_init(sync_block_1, module_dict, config)

    # Add 2nd Sync Multi-Sequence Block to the Sync While sequence
    sync_block_2 = sync_sequence.add_sync_multi_sequence_block("Mixed Signal Generation", 10)
    # Program Mixed Signal Generation
    program_mixed_sig_gen(sync_block_2, module_dict, config)


def program_loop_init(sync_block, module_dict, config):
    "Programs the Sync Block that waits for an external trigger and initializes the RF Gen"
    # Program digitizer sequence to wait for ext. trigger
    dig_sequence = sync_block.sequences[config.dig_engine]
    program_wait_trigger(dig_sequence, config)
    
    # Program RF Gen. Initialization
    rf_gen_sequence = sync_block.sequences[config.rf_gen_engine]
    rf_gen_module = module_dict[config.rf_gen_engine].instrument
    program_rf_gen_init(rf_gen_sequence, rf_gen_module, config)


def program_wait_trigger(dig_sequence, config):
    "Wait for Trigger and increment loop counter"
    # Previously defined registers
    iteration_counter = dig_sequence.scope.registers[config.iteration_counter]
    
    # Define the condition for the wait statement
    trigger_event = dig_sequence.engine.triggers[config.fp_trigger]
    wait_condition = kthvi.Condition.trigger(trigger_event)
    # Add a Wait For Event
    wait_event = dig_sequence.add_wait("Wait for FP Trigger", 10, wait_condition)
    wait_event.set_mode(kthvi.WaitMode.TRANSITION, kthvi.SyncMode.IMMEDIATE)

    # Increment the sync while iteration counter for each external trigger event that is received
    instruction = dig_sequence.add_instruction("Increment Counter", 50, dig_sequence.instruction_set.add.id)
    instruction.set_parameter(dig_sequence.instruction_set.add.destination.id, iteration_counter)
    instruction.set_parameter(dig_sequence.instruction_set.add.left_operand.id, iteration_counter)
    instruction.set_parameter(dig_sequence.instruction_set.add.right_operand.id, 1)


def program_rf_gen_init(rf_gen_sequence, rf_gen_module, config):
    "Initialize the RF Generator"

    # Previously defined registers
    pulse_counter = rf_gen_sequence.scope.registers[config.pulse_counter]

    # Initialize pulse_counter = 0
    instruction = rf_gen_sequence.add_instruction("Initialize Pulse Counter", 10, rf_gen_sequence.instruction_set.assign.id)
    instruction.set_parameter(rf_gen_sequence.instruction_set.assign.destination.id, pulse_counter)
    instruction.set_parameter(rf_gen_sequence.instruction_set.assign.source.id, 0)

    # Set CH1 Frequency
    instruction = rf_gen_sequence.add_instruction("Set CH1 Frequency", 10, rf_gen_module.hvi.instruction_set.set_frequency.id)
    instruction.set_parameter(rf_gen_module.hvi.instruction_set.set_frequency.channel.id, config.n_AWG)
    instruction.set_parameter(rf_gen_module.hvi.instruction_set.set_frequency.value.id, config.frequency)


def program_mixed_sig_gen(sync_block, module_dict, config):
    "Programs the Mixed Signal Generation Sync Multi-Sequence Block"

    # Digitizer sequence: DAQ trigger
    dig_sequence = sync_block.sequences[config.dig_engine]
    program_daq_trigger(dig_sequence)

    # RF Gen. sequence: genearete RF pulses
    rf_gen_sequence = sync_block.sequences[config.rf_gen_engine]
    rf_gen_module = module_dict[config.rf_gen_engine].instrument
    program_rf_pulses(rf_gen_sequence, rf_gen_module, config)

    # Sequence of AWG 2:N: queue and play waveforms
    for engine_name in module_dict.keys():
        if engine_name!=config.rf_gen_engine and engine_name!=config.dig_engine:
            awg_sequence = sync_block.sequences[engine_name]
            awg_module = module_dict[engine_name]
            program_queue_play_wfm(awg_sequence, awg_module, config)


def program_daq_trigger(dig_sequence):
    "Executes all the previously added DAQ trigger actions"

    # Action-execute
    daq_trigger_all = dig_sequence.engine.actions
    inst_daq_trigger = dig_sequence.add_instruction("DAQ Trigger", 10, dig_sequence.instruction_set.action_execute.id)
    inst_daq_trigger.set_parameter(dig_sequence.instruction_set.action_execute.action.id, daq_trigger_all)


def program_rf_pulses(rf_gen_sequence, rf_gen_module, config):
    "Programs RF pulse generation"

    # Previously defined registers
    pulse_counter = rf_gen_sequence.scope.registers[config.pulse_counter]
    
    # Local WHILE: amplitude ON, amplitude OFF
    local_while_condition = kthvi.Condition.register_comparison(pulse_counter, kthvi.ComparisonOperator.LESS_THAN, config.num_pulses)
    while_loop = rf_gen_sequence.add_while("RF pulses", 70, local_while_condition)
    while_sequence = while_loop.sequence
    #Increment pulse_counter
    instruction = while_sequence.add_instruction("Increment Pulse Counter", 100, while_sequence.instruction_set.add.id)
    instruction.set_parameter(while_sequence.instruction_set.add.destination.id, pulse_counter)
    instruction.set_parameter(while_sequence.instruction_set.add.left_operand.id, pulse_counter)
    instruction.set_parameter(while_sequence.instruction_set.add.right_operand.id, 1)
    # Set CH1 amplitude to ON_value
    instruction = while_sequence.add_instruction("Set CH1 amplitude to ON_value", config.pulse_delay, rf_gen_module.hvi.instruction_set.set_amplitude.id)
    instruction.set_parameter(rf_gen_module.hvi.instruction_set.set_amplitude.channel.id, config.n_AWG)
    instruction.set_parameter(rf_gen_module.hvi.instruction_set.set_amplitude.value.id, config.ON_value)
    # Set CH1 amplitude to OFF_value
    # NOTE: the RF pulse ON time value is implemented using a Python variable because it is not expected to change during HVI execution
    # To change the RF pulse on time during HVI execution you can use a WaitTime statement instead of the Python variable
    instruction = while_sequence.add_instruction("Set CH1 amplitude to OFF_value", config.pulse_ontime, rf_gen_module.hvi.instruction_set.set_amplitude.id)
    instruction.set_parameter(rf_gen_module.hvi.instruction_set.set_amplitude.channel.id, config.n_AWG)
    instruction.set_parameter(rf_gen_module.hvi.instruction_set.set_amplitude.value.id, config.OFF_value)


def program_queue_play_wfm(awg_sequence, awg_module, config):
    "Programs AWG sequence to queue and play wfms"

    # Recall registers
    queue_reg = awg_sequence.scope.registers[config.queue_reg]
    reg_wfm_A = awg_sequence.scope.registers[config.reg_wfm_A]
    reg_wfm_B = awg_sequence.scope.registers[config.reg_wfm_B]

    # If branch
    if_condition = kthvi.Condition.register_comparison(queue_reg, kthvi.ComparisonOperator.EQUAL_TO, 0)
    enable_ifbranches_time_matching = True
    if_statement = awg_sequence.add_if("Queue Wfm AWG", 70, if_condition, enable_ifbranches_time_matching)
    if_branch_seq = if_statement.if_branch.sequence
    for awg_ch in range(1, awg_module.num_channels+1):
        instrLabel = "Queue Waveform A CH" + str(awg_ch)
        instruction = if_branch_seq.add_instruction(instrLabel, config.queue_wfm_latency, awg_module.instrument.hvi.instruction_set.queue_waveform.id)
        #Set every parameter of AWGqueueWaveform(awg_ch, waveformNumber, triggerMode, startDelay, cycles, prescaler);
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.waveform_number.id, reg_wfm_A)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.channel.id, awg_ch)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.trigger_mode.id, config.awg_trigger_mode)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.start_delay.id, config.start_delay)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.cycles.id, config.wfm_A_cycles)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.prescaler.id, config.awg_prescaler)
    # Else branch
    else_branch_seq = if_statement.else_branch.sequence
    for awg_ch in range(1, awg_module.num_channels+1):
        instrLabel = "Queue Waveform B CH" + str(awg_ch)
        instruction = else_branch_seq.add_instruction(instrLabel, config.queue_wfm_latency, awg_module.instrument.hvi.instruction_set.queue_waveform.id)
        #Set every parameter of AWGqueueWaveform(awg_ch, waveformNumber, triggerMode, startDelay, cycles, prescaler);
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.waveform_number.id, reg_wfm_B)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.channel.id, awg_ch)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.trigger_mode.id, config.awg_trigger_mode)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.start_delay.id, config.start_delay)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.cycles.id, config.wfm_B_cycles)
        instruction.set_parameter(awg_module.instrument.hvi.instruction_set.queue_waveform.prescaler.id, config.awg_prescaler)

    # AWG trigger
    action_list = awg_sequence.engine.actions
    inst_awg_trigger = awg_sequence.add_instruction("AWG Trigger", config.awg_trigger_latency, awg_sequence.instruction_set.action_execute.id)
    inst_awg_trigger.set_parameter(awg_sequence.instruction_set.action_execute.action.id, action_list)


def program_execution_completed(sync_block, config):
    "Set hvi_done = 1 to complete the execution"

    # Set hvi_done = 1 in digitizer module
    sequence = sync_block.sequences[config.dig_engine]
    reg_hvi_done = sequence.scope.registers[config.hvi_done]
    instruction = sequence.add_instruction("hvi_done = 1", 10, sequence.instruction_set.assign.id)
    instruction.set_parameter(sequence.instruction_set.assign.destination.id, reg_hvi_done)
    instruction.set_parameter(sequence.instruction_set.assign.source.id, 1)


def choose_wfm(hvi, module_dict, config):
    "While HVI executes user can change the waveform that AWGs 2:N are playing"
    
    print("Press: 1 to Queue'n'Play Wfm A, 2 to Queue'n'Play Wfm B, Enter to continue")
    key_input = input()
    if key_input == '1':
        for engine_name in module_dict.keys():
            if engine_name!=config.rf_gen_engine and engine_name!=config.dig_engine:
                hvi.sync_sequence.scopes[engine_name].registers[config.queue_reg].write(0)
    if key_input == '2':
        for engine_name in module_dict.keys():
            if engine_name!=config.rf_gen_engine and engine_name!=config.dig_engine:
                hvi.sync_sequence.scopes[engine_name].registers[config.queue_reg].write(1)


def export_hvi_seqeunces(sequencer):
    """
    Exports the programmed HVI sequences to text format
    """
    # Generate HVI sequence description text
    output = sequencer.sync_sequence.to_string(kthvi.OutputFormat.DEBUG)
    print("Programmed HVI sequences exported to file")
    
    # Write file
    text_file = open("TextOutput/HviMixedSigMeas.txt", "w")
    text_file.write(output)
    text_file.close()


def interact_with_hvi(hvi, module_dict, config):
    """
    This function allows users to interact with the HVI execution from SW execution
    """ 
    # Load configuration
    digitizer_module = module_dict[config.dig_engine].instrument
    
    # Access registers from the compiled HVI to read/write them during HVI execution 
    iteration_counter = hvi.sync_sequence.scopes[config.dig_engine].registers[config.iteration_counter]
    pulse_counter = hvi.sync_sequence.scopes[config.rf_gen_engine].registers[config.pulse_counter]
    queue_reg = hvi.sync_sequence.scopes[config.awg_engine + "0"].registers[config.queue_reg]
    hvi_done = hvi.sync_sequence.scopes[config.dig_engine].registers[config.hvi_done]

    print("\n####### Application Parameters #######")
    print("Number of HVI loops: {}".format(config.num_loops))
    print("Digitizer configured to acquire {} points in {} cycles".format(config.acquisition_points_per_cycle, config.dig_num_cycles))
    print("Total dig. acquisition_points = points_per_cycle*num_cycles = {}".format(config.acquisition_points))
    print("Number of RF Pulses: {}".format(config.num_pulses))
    print("RF Pulse ON Time: {} ns".format(config.pulse_ontime))
    print("RF pulses frequency: {} Mhz".format(int(config.frequency/1e6)))
    print("AWG waveforms loaded to RAM: 2")
    # User can choose the waveform to be played while HVI is running
    choose_wfm(hvi, module_dict, config)
    print()
    print("######################################")
    print("Please connect your External Trigger source to the FP TRIG_OUT connector of DIG module in chassis {}, slot {}".format(config.digitizer_descriptor.chassis_number, config.digitizer_descriptor.slot_number))
    print("Once the FP trigger source is connected please run {} FP trigger events for the HVI sequences execution to complete".format(config.num_loops))
    print("######################################")
    print("Waiting for FP trigger...")
    print()
    
    # User controlled loop
    iterations = 0
    attempts = 0
    max_attempts = 50
    while (iterations < config.num_loops) and (attempts<max_attempts):
        time.sleep(1) #waits for 1s
        iterations = iteration_counter.read()
        attempts += 1
    
    if attempts>=max_attempts:
        print("Timeout.. Quitting the application execution...")

    if hvi_done.read() == 1:
        print("HVI Execution Completed Successfully!")
        # Digitizer acquisition
        timeout = 5000 #[ms]
        measured_data1 = numpy.zeros(config.acquisition_points)
        measured_data1 = digitizer_module.DAQread(1, config.acquisition_points, timeout)
        measured_data2 = numpy.zeros(config.acquisition_points)
        measured_data2 = digitizer_module.DAQread(2, config.acquisition_points, timeout)
        # Plot acquisition data
        plt.clf()
        plt.plot(measured_data1, 'b-', measured_data2, 'r-')
        plt.show()
    else:
        print("HVI Execution Failed!")

    # Summary of register values
    print("\n=== Final Register Values ===")
    print("Iteration counter: {}".format(iteration_counter.read()))
    print("Pulse counter: {}".format(pulse_counter.read()))
    print("Queue Reg: {}".format(queue_reg.read()))
    print("HVI Done: {}".format(hvi_done.read()))
    print()

######################################################
# 3rd Level: Functions to Use SD1/M3xxxA Instruments
######################################################

class ModuleDescriptor:
    "Descriptor for module objects"
    def __init__(self, model_number, chassis_number, slot_number, options, engine_name):
        self.model_number = model_number
        self.chassis_number = chassis_number
        self.slot_number = slot_number
        self.options = options
        self.engine_name = engine_name


class M9031Descriptor:
    "Describes the interconnection between each pair of M9031A modules"
    def __init__(self, first_M9031_chassis_number, first_M9031_slot_number, second_M9031_chassis_number, second_M9031_slot_number):
        self.chassis_1 = first_M9031_chassis_number
        self.slot_1 = first_M9031_slot_number
        self.chassis_2 = second_M9031_chassis_number
        self.slot_2 = second_M9031_slot_number


class Module:
    "Class defining a modular instrument object and its properties"
    def __init__(self, instrument_object, num_channels):
        self.instrument = instrument_object
        self.num_channels = num_channels


def open_modules(config):
    """ 
    Opens and creates all the necessary instrument objects.
    Returns a dictionary of Module objects whose keys are the HVI engine names.
    Please check SD1 3.x User Manual for options to open the instrument objects
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Open SD1 instrument objects
    for descriptor in config.module_descriptors:
        if descriptor.model_number == "M3102A":
            instr_obj = keysightSD1.SD_AIN()
        else:
            instr_obj = keysightSD1.SD_AOU()
        instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
        id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
        if id < 0:
            raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {} - {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
        nCh = instr_obj.getOptions("channels")
        if nCh == "CH2":
            num_channels = 2
        elif nCh == "CH4":
            num_channels = 4
        else:
            raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        module_dict.update( {descriptor.engine_name : Module(instr_obj, num_channels)} )
        num_modules += 1
    
    return module_dict


def configure_awg(config, awg_module, num_channels = 2):
    """ 
    configure_awg(awg_module, num_channels) configures an awg_module to output arbitrary waveforms.
    Waveforms are going to be queued and triggered from the HVI sequence.
    """
    # Load waveform to AWG memory
    awg_module.waveformFlush() #memory flush
    wave = keysightSD1.SD_Wave()
    wave.newFromFile("Waveforms/Gaussian.csv")
    awg_module.waveformLoad(wave, config.wfm_A)
    wave.newFromFile("Waveforms/Exponential.csv")
    awg_module.waveformLoad(wave, config.wfm_B)

    for awg_ch in range(1, num_channels+1):
        # AWG queue flush 
        awg_module.AWGstop(awg_ch)
        awg_module.AWGflush(awg_ch)
        # Set AWG mode
        awg_module.channelWaveShape(awg_ch, config.awg_mode)
        awg_module.channelAmplitude(awg_ch, config.amplitude)
        # AWG configuration
        awg_module.AWGqueueConfig(awg_ch, config.queue_mode)
        awg_module.AWGqueueSyncMode(awg_ch, config.sync_mode)
        # Start AWG
        awg_module.AWGstart(awg_ch) # AWG starts and waits for an AWG trigger


def configure_rfgen(config, awg_module, num_channels = 2):
    """
    This method configures AWG channels to work as RF pulse generators
    Note: Keysight channel numbering starts from 1
    """
    # Setup AWG to work as a sine wave generator
    for awg_ch in range(1, num_channels+1):
        awg_module.channelWaveShape(awg_ch, keysightSD1.SD_Waveshapes.AOU_SINUSOIDAL)
        awg_module.channelFrequency(awg_ch, config.frequency)
        awg_module.channelOffset(awg_ch, config.offset)
        awg_module.channelAmplitude(awg_ch, config.OFF_value)
        awg_module.channelPhaseResetMultiple(config.all_ch_mask)


def configure_digitizer(config, digitizer_module, num_channels = 4):
    """ 
    This method configures multiple DAQ channels for cyclic acquisitions
    """
    # Input settings
    prescaler = config.dig_prescaler 
    fullscale = config.fullscale
    num_cycles = config.dig_num_cycles
    points_per_cycle = config.acquisition_points_per_cycle
    acquisition_delay = config.acquisition_delay
    trigger_mode = config.dig_trigger_mode

    # Set the FP TRG_OUT port to be used as an input for the ext. trigger event
    error = digitizer_module.triggerIOconfig(keysightSD1.SD_TriggerDirections.AOU_TRG_IN)
    if error < 0:
        raise Exception("Error setting the digitizer TRG port! Error code: {} - {}. Exiting...".format(error, keysightSD1.SD_Error.getErrorMessage(error)))

    # DAQstop: stops any previous acquisition
    digitizer_module.DAQstopMultiple(config.mask)
    
    # Configure DAQ channels
    for n_DAQ in range (1, num_channels+1):
        # Resets the DAQ
        digitizer_module.DAQflush(n_DAQ)
        # Digitizer channel configuration
        digitizer_module.channelInputConfig(n_DAQ, fullscale, keysightSD1.AIN_Impedance.AIN_IMPEDANCE_50, keysightSD1.AIN_Coupling.AIN_COUPLING_DC)
        digitizer_module.channelPrescalerConfig(n_DAQ, prescaler)
        # DAQ acquisitions configuration
        digitizer_module.DAQconfig(n_DAQ, points_per_cycle, num_cycles, acquisition_delay, trigger_mode)
    
    #DAQstart: digitizer channels get strted and are waiting for the DAQ trigger commands that trigger the acquisition of each cycle of data points
    digitizer_module.DAQstartMultiple(config.mask)
    
    return points_per_cycle


if __name__ == '__main__':
    main()